"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.getAuthUser = exports.createCustomToken = exports.callFirestore = exports.callRtdb = void 0;
const database_1 = require("firebase-admin/database");
const auth_1 = require("firebase-admin/auth");
const firestore_1 = require("firebase-admin/firestore");
const app_1 = require("firebase-admin/app");
const firebase_utils_1 = require("./firebase-utils");
/**
 * @param baseRef - Base RTDB reference
 * @param options - Options for ref
 * @returns RTDB Reference
 */
function optionsToRtdbRef(baseRef, options) {
    let newRef = baseRef;
    [
        'orderByChild',
        'orderByKey',
        'orderByValue',
        'equalTo',
        'startAfter',
        'startAt',
        'endBefore',
        'endAt',
        'limitToFirst',
        'limitToLast',
    ].forEach((optionName) => {
        if (options && options[optionName]) {
            const args = options[optionName];
            // Spread arg arrays (such as startAfter and endBefore)
            if (Array.isArray(args)) {
                newRef = newRef[optionName](...args);
            }
            else {
                newRef = newRef[optionName](args);
            }
        }
    });
    return newRef;
}
/**
 * Get Firebase Auth or TenantAwareAuth instance, based on tenantId being provided
 * @param authSettings - Optional ID of tenant used for multi-tenancy
 * @param authSettings.tenantId - Optional ID of tenant used for multi-tenancy
 * @param authSettings.appName - Optional name of Firebase app. Defaults to "[DEFAULT]"
 * @returns Firebase Auth or TenantAwareAuth instance
 */
function getAdminAuthWithTenantId(authSettings) {
    const { tenantId, appName } = authSettings || {};
    const authInstance = (0, auth_1.getAuth)(appName ? (0, app_1.getApp)(appName) : undefined);
    const auth = tenantId
        ? authInstance.tenantManager().authForTenant(tenantId)
        : authInstance;
    return auth;
}
/**
 * Convert unique data types which have been stringified and parsed back
 * into their original type.
 * @param dataVal - Value of data
 * @returns Value converted into timestamp object if possible
 */
function convertValueToTimestampOrGeoPointIfPossible(dataVal) {
    /* eslint-disable no-underscore-dangle */
    if (dataVal?._methodName === 'serverTimestamp' ||
        dataVal?._methodName === 'FieldValue.serverTimestamp' // v8 and earlier
    ) {
        return firestore_1.FieldValue.serverTimestamp();
    }
    if (dataVal?._methodName === 'deleteField' ||
        dataVal?._methodName === 'FieldValue.delete' // v8 and earlier
    ) {
        return firestore_1.FieldValue.delete();
    }
    /* eslint-enable no-underscore-dangle */
    if (typeof dataVal?.seconds === 'number' &&
        typeof dataVal?.nanoseconds === 'number') {
        return new firestore_1.Timestamp(dataVal.seconds, dataVal.nanoseconds);
    }
    if (typeof dataVal?.latitude === 'number' &&
        typeof dataVal?.longitude === 'number') {
        return new firestore_1.GeoPoint(dataVal.latitude, dataVal.longitude);
    }
    return dataVal;
}
/**
 * @param data - Data to be set in firestore
 * @returns Data to be set in firestore with timestamp
 */
function getDataWithTimestampsAndGeoPoints(data) {
    return Object.entries(data).reduce((acc, [currKey, currData]) => {
        // Convert nested timestamp if item is an object
        if (typeof currData === 'object' &&
            currData !== null &&
            !Array.isArray(currData) &&
            /* eslint-disable-next-line no-underscore-dangle */
            !currData._methodName &&
            !currData.seconds &&
            !(currData.latitude && currData.longitude)) {
            return {
                ...acc,
                [currKey]: getDataWithTimestampsAndGeoPoints(currData),
            };
        }
        const value = Array.isArray(currData)
            ? currData.map((dataItem) => {
                const result = convertValueToTimestampOrGeoPointIfPossible(dataItem);
                return result.constructor === Object
                    ? getDataWithTimestampsAndGeoPoints(result)
                    : result;
            })
            : convertValueToTimestampOrGeoPointIfPossible(currData);
        return {
            ...acc,
            [currKey]: value,
        };
    }, {});
}
/**
 * @param action - Action to run
 * @param actionPath - Path in RTDB
 * @param options - Query options
 * @param data - Data to pass to action
 * @returns Promise which resolves with results of calling RTDB
 */
async function callRtdb(action, actionPath, options, data) {
    // Handle actionPath not being set (see #244 for more info)
    if (!actionPath) {
        throw new Error('actionPath is required for callRtdb. Use "/" for top level actions.');
    }
    try {
        const dbInstance = (0, database_1.getDatabase)(options?.appName ? (0, app_1.getApp)(options.appName) : undefined);
        const dbRef = dbInstance.ref(actionPath);
        if (action === 'get') {
            const snap = await optionsToRtdbRef(dbRef, options).once('value');
            return snap.val();
        }
        if (action === 'push') {
            const pushRef = dbRef.push();
            await pushRef.set(data);
            // TODO: Return key on an object for consistent return regardless of action
            return pushRef.key;
        }
        // Delete action
        const actionNameMap = {
            delete: 'remove',
        };
        const cleanedActionName = actionNameMap[action] || action;
        await dbRef[cleanedActionName](data);
        // Prevents Cypress error with message:
        // "You must return a promise, a value, or null to indicate that the task was handled."
        return null;
    }
    catch (err) {
        /* eslint-disable no-console */
        console.error(`cypress-firebase: Error with RTDB "${action}" at path "${actionPath}" :`, err);
        /* eslint-enable no-console */
        throw err;
    }
}
exports.callRtdb = callRtdb;
/**
 * @param action - Action to run
 * @param actionPath - Path to collection or document within Firestore
 * @param options - Query options
 * @param data - Data to pass to action
 * @returns Promise which resolves with results of calling Firestore
 */
async function callFirestore(action, actionPath, options, data) {
    const firestoreInstance = (0, firestore_1.getFirestore)((0, app_1.getApp)(options?.appName));
    try {
        if (action === 'get') {
            const snap = await (0, firebase_utils_1.slashPathToFirestoreRef)(firestoreInstance, actionPath, options).get();
            if (snap?.docs?.length && typeof snap.docs.map === 'function') {
                return snap.docs.map((docSnap) => ({
                    ...docSnap.data(),
                    id: docSnap.id,
                }));
            }
            // Falling back to null in the case of falsey value prevents Cypress error with message:
            // "You must return a promise, a value, or null to indicate that the task was handled."
            return (typeof snap?.data === 'function' && snap.data()) || null;
        }
        if (action === 'delete') {
            // Handle deleting of collections & sub-collections if not a doc path
            const deletePromise = (0, firebase_utils_1.isDocPath)(actionPath)
                ? (0, firebase_utils_1.slashPathToFirestoreRef)(firestoreInstance, actionPath, options).delete()
                : (0, firebase_utils_1.deleteCollection)(firestoreInstance, (0, firebase_utils_1.slashPathToFirestoreRef)(firestoreInstance, actionPath, options), options);
            await deletePromise;
            // Returning null in the case of falsey value prevents Cypress error with message:
            // "You must return a promise, a value, or null to indicate that the task was handled."
            return null;
        }
        if (!data) {
            throw new Error(`You must define data to run ${action} in firestore.`);
        }
        const dataToSet = getDataWithTimestampsAndGeoPoints(data);
        if (action === 'set') {
            return firestoreInstance
                .doc(actionPath)
                .set(dataToSet, options?.merge
                ? { merge: options?.merge }
                : undefined);
        }
        // "update" action
        return (0, firebase_utils_1.slashPathToFirestoreRef)(firestoreInstance, actionPath, options)[action](dataToSet);
    }
    catch (err) {
        /* eslint-disable no-console */
        console.error(`cypress-firebase: Error with Firestore "${action}" at path "${actionPath}" :`, err);
        /* eslint-enable no-console */
        throw err;
    }
}
exports.callFirestore = callFirestore;
/**
 * Create a custom token
 * @param settings - Settings object
 * @returns Promise which resolves with a custom Firebase Auth token
 */
function createCustomToken(settings) {
    // Use custom claims or default to { isTesting: true }
    const customClaims = settings?.customClaims || { isTesting: true };
    // Create auth token
    return getAdminAuthWithTenantId(settings).createCustomToken(settings.uid, customClaims);
}
exports.createCustomToken = createCustomToken;
/**
 * Get Firebase Auth user based on UID
 * @param settings - Task settings
 * @param settings.uid - UID of user for which the custom token will be generated
 * @param settings.tenantId - Optional ID of tenant used for multi-tenancy
 * @returns Promise which resolves with a custom Firebase Auth token
 */
function getAuthUser(settings) {
    return getAdminAuthWithTenantId(settings).getUser(settings.uid);
}
exports.getAuthUser = getAuthUser;
